# -*- coding: utf-8 -*-

from django.conf import settings
from django.contrib.auth.models import User, UserManager
from django.core.exceptions import SuspiciousOperation
from django.urls import reverse
from django.db import models
from django.db.models import signals
from django.db.models.base import ModelBase
from django.http import Http404
from django.utils.safestring import mark_safe

from datetime import date, datetime, timedelta
from hashlib import sha1
from os import environ
from uuid import uuid4

from ldap_inp import LdapEleve, LdapEcole
from edt.models import EdtAdeGroup, EdtRecurring


class EcoleMetaclass(ModelBase):
    def __new__(mcs, name, bases, dict):
        klass = ModelBase.__new__(mcs, name, bases, dict)

        klass.inp = klass(id=1, ident="inp", nom_propre="INP Toulouse", nom_commun="INP")
        klass.n7 = klass(id=2, ident="n7", nom_propre="ENSEEIHT", nom_commun="n7", autoinscription=True)
        klass.ensat = klass(id=3, ident="ensat", nom_propre="ENSAT", nom_commun="ENSAT", autoinscription=True)
        klass.a7 = klass(id=4, ident="a7", nom_propre="ENSIACET", nom_commun="A7", autoinscription=True)
        klass.enm = klass(id=5, ident="enm", nom_propre="ENM", nom_commun="ENM", ldap_cri=False)

        return klass


class Ecole(models.Model, metaclass=EcoleMetaclass):
    ident = models.CharField(max_length=5, unique=True)
    nom_propre = models.CharField(max_length=15, unique=True)
    nom_commun = models.CharField(max_length=15, unique=True)
    # indique si le LDAP du CRI doit être utilisé pour valider les inscriptions
    ldap_cri = models.BooleanField()
    # indique si les inscriptions directes par les utilisateurs sont autorisées
    autoinscription = models.BooleanField()

    def __init__(self, id, ident, nom_propre, nom_commun, ldap_cri=True, autoinscription=False):
        models.Model.__init__(
            self,
            id=id,
            ident=ident,
            nom_propre=nom_propre,
            nom_commun=nom_commun,
            ldap_cri=ldap_cri,
            autoinscription=autoinscription,
        )

        self.dn = getattr(LdapEcole, ident).dn

    @staticmethod
    def creationInit(sender, **kwargs):
        models.Model.save(Ecole.inp)
        models.Model.save(Ecole.n7)
        models.Model.save(Ecole.ensat)
        models.Model.save(Ecole.a7)
        models.Model.save(Ecole.enm)

    @staticmethod
    def ecoleActuelle(request):
        if settings.ECOLE_PAR_ENV:
            return getattr(Ecole, environ["ECOLE"])

        try:
            ident = settings.CORRESP_HOTES[request.get_host()]
        except KeyError:
            ident = "inp"

        return getattr(Ecole, ident)

    def delete(self, *args, **kwargs):
        raise SuspiciousOperation("La modification des écoles n’est pas autorisée")

    def save(self, *args, **kwargs):
        raise SuspiciousOperation("La modification des écoles n’est pas autorisée")

    def get_absolute_url(self):
        return "//" + settings.URLS_SITES[self.id - 1]

    def __eq__(self, other):
        try:
            return self.dn == other.dn
        except AttributeError:
            return models.Model.__eq__(self, other)

    def __ne__(self, other):
        try:
            return self.dn != other.dn
        except AttributeError:
            return models.Model.__ne__(self, other)

    def __str__(self):
        return self.nom_commun


class Permission(models.Model):
    ident = models.CharField(max_length=64)
    nom = models.CharField(max_length=25)
    desc = models.CharField(max_length=128)
    ecole = models.ForeignKey(Ecole, null=True, on_delete=models.CASCADE)

    def delete(self):
        raise SuspiciousOperation("La modification des permissions n’est pas autorisée")

    def save(self):
        raise SuspiciousOperation("La modification des permissions n’est pas autorisée")

    def __str__(self):
        return "{}, ecole = {}".format(self.ident, str(self.ecole))

    class Meta:
        unique_together = (("ident", "ecole"),)


class PortailUserManager(UserManager):
    def create_user(self, username, email, ecole, password=None):
        """
        Creates and saves a User with the given username, e-mail and password.
        """

        now = datetime.now()

        # Normalize the address by lowercasing the domain part of the email
        # address.
        try:
            email_name, domain_part = email.strip().split("@", 1)
        except ValueError:
            pass
        else:
            email = "@".join([email_name, domain_part.lower()])

        user = self.model(
            username=username,
            email=email,
            is_staff=False,
            is_active=True,
            is_superuser=False,
            last_login=now,
            date_joined=now,
            ecole=ecole,
        )

        if password:
            user.set_password(password)
        else:
            user.set_unusable_password()
        user.save(using=self._db)
        return user

    def create_superuser(self, username, email, ecole, password):
        u = self.create_user(username, email, ecole, password)
        u.is_staff = False
        u.is_active = True
        u.is_superuser = True
        u.save(using=self._db)
        return u

    def create_from_ldap(self, ldap_user: LdapEleve, parent_user: User = None):
        """
        Crée un nouvel utilisateur dans la DB SQL, ou
        attache un utilisateur à un object auth.User.existant.
        """

        mail = ldap_user.mail
        nom = ldap_user.sn
        prenom = ldap_user.givenName
        ecole = Ecole.objects.get(ident=ldap_user.get_ecole())

        args = {
            "username": ldap_user.uid.lower(),
            "email": mail,
            "first_name": prenom,
            "last_name": nom,
            "ecole": ecole,
        }

        if parent_user is not None:
            args["user_ptr"] = parent_user

        user, _ = PortailUser.objects.get_or_create(**args)

        user.force_change_passwd = True
        user.set_unusable_password()

        return user

    def get_from_ldap(self, username):
        if not username.isalnum():
            raise PortailUser.DoesNotExist("Incorrect username " + username)

        try:
            return self.get(username=username.lower())
        except PortailUser.DoesNotExist as e:
            try:
                ldap_user = LdapEleve.objects.get(uid=username)
            except LdapEleve.DoesNotExist:
                raise e
            user = self.create_from_ldap(ldap_user)
            user.is_superuser = user.ldap_is_superuser()
            user.is_staff = user.ldap_is_staff()
            user.save()
            return user

    def get_from_ldap_or_404(self, username):
        try:
            return self.get_from_ldap(username)
        except PortailUser.DoesNotExist:
            raise Http404


class PortailUser(User):
    ecole = models.ForeignKey("Ecole", on_delete=models.CASCADE)
    permissions = models.ManyToManyField(Permission)

    edt_ade_groups = models.ManyToManyField(EdtAdeGroup)
    edtrecurring = models.ManyToManyField(EdtRecurring)
    edt_club_events = models.SmallIntegerField(null=True)
    force_change_passwd = models.BooleanField(default=False)

    objects = PortailUserManager()

    def check_permission(self, request, permission):
        if self.is_anonymous:
            return False

        if self.is_superuser or (self.is_staff and self.ecole == request.ecole):
            return True

        if not hasattr(self, "_permissions_cache"):
            self._permissions_cache = {}
            for ecole in Ecole.objects.all():
                self._permissions_cache[ecole.pk] = [perm.ident for perm in self.permissions.filter(ecole=ecole)]
            self._permissions_cache[0] = [perm.ident for perm in self.permissions.filter(ecole=None)]

        return permission in self._permissions_cache[request.ecole.pk] or permission in self._permissions_cache[0]

    def get_ldap_user(self):
        return LdapEleve.objects.get(uid=self.username)

    def get_nickname(self):
        nickname = self.get_ldap_user().nickname

        if not nickname:
            nickname = self.get_full_name()

        return nickname

    def get_is_masculin(self):
        return self.get_ldap_user().genre == 1

    def get_link(self):
        url = reverse("annuaire:fiche", args=[self.username])
        return mark_safe('<a href="{}">{}</a>'.format(url, self.get_full_name()))

    def get_absolute_url(self):
        return reverse("annuaire:fiche", args=[self.username])

    def get_dn(self):
        return self.get_ldap_user().dn

    def get_promo(self):
        return self.get_ldap_user().promo

    def get_mobile(self):
        return self.get_ldap_user().mobile or ""

    def get_formated_mobile(self):
        return self.format_tel(self.get_mobile())

    def get_mail_ecole(self):
        return self.get_ldap_user().mailEcole

    def get_groups(self, ecole=None):
        return self.get_ldap_user().get_groupes()

    def get_clubs(self, ecole=None):
        return self.get_ldap_user().get_clubs(ecole)

    def ldap_is_superuser(self):
        return self.get_ldap_user().is_superuser()

    def ldap_is_staff(self):
        return self.get_ldap_user().is_staff()

    @property
    def has_configured_timetables(self):
        return self.edt_club_events is not None

    @staticmethod
    def format_tel(value, lien=True):
        """
        Le numéro de tel ``value`` sera préfixé par 0, pas par +33
        """
        if len(value) == 12 and value[:3] == "+33":
            num = " ".join("".join(x) for x in zip("0" + value[4::2], value[3::2]))
        else:
            num = value

        if value and lien:
            return mark_safe('<a href="tel:{}">{}</a>'.format(value, num))

        return value

    @staticmethod
    def extended_promo(promo):
        """
        Depuis l'année de sortie d'école ``promo``, renvoie l'année
        d'étude.
        """
        today = date.today()
        if today < date(year=today.year, month=9, day=1):
            # printemps
            anneeA = today.year - promo + 3
        else:
            # automne
            anneeA = today.year - promo + 3 + 1

        return anneeA


def valeur_default():
    return sha1(uuid4().bytes).hexdigest()


def date_limite_default():
    return datetime.now() + timedelta(minutes=1)


class ApplicationLog(models.Model):
    name = models.SlugField(max_length=32, unique=True)

    def __str__(self):
        return self.name


class LogEntry(models.Model):
    application = models.ForeignKey(ApplicationLog, on_delete=models.CASCADE)
    date = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(PortailUser, related_name="logs", on_delete=models.CASCADE)
    operation = models.TextField()

    def __str__(self):
        return "{} {} {}: {}".format(self.application, self.date, self.user, self.operation)


signals.post_migrate.connect(Ecole.creationInit, dispatch_uid="portailuser.models.Ecole.creationInit")
